"""Automatic generation control models"""

from andes.core.param import NumParam, ExtParam  # NOQA
from andes.core.model import Model, ModelData  # NOQA


class AGCControlData(ModelData):
    pass


class AGCControl:
    pass
